`include "xgriscv_defines.v"

module forwarding(
	input			pcsrcID, memwriteID, branchID, jalrID,
					regwriteEX, memtoregEX, memwriteEX, regwriteMEM, memtoregMEM, memwriteMEM, regwriteWB, memtoregWB,
	input [4:0]		rs1ID, rs2ID, rs1EX, rs2EX, rdEX, rs2MEM, rdMEM, rdWB,
	output			forwardaID, forwardbID,
	output[1:0] 	forwardaEX, forwardbEX,
	output			forwardMEM,
	output			stallIF, stallID, flushID, flushEX);

	//for beq in the ID stage
	assign forwardaID = regwriteMEM & (rdMEM != 0) & ~memtoregMEM & (rdMEM == rs1ID);
	assign forwardbID = regwriteMEM & (rdMEM != 0) & ~memtoregMEM & (rdMEM == rs2ID);
	
	//for ops in the EX stage
	wire forwardaEXfromMEM = regwriteMEM & (rdMEM != 0) & (rdMEM == rs1EX);
	wire forwardaEXfromWB = !forwardaEXfromMEM & regwriteWB & (rdWB != 0) & (rdWB == rs1EX);
	assign forwardaEX = ({2{forwardaEXfromMEM}} & 2'b10) | ({2{forwardaEXfromWB}} & 2'b01);

	wire forwardbEXfromMEM = regwriteMEM & (rdMEM != 0) & (rdMEM == rs2EX);
	wire forwardbEXfromWB = !forwardbEXfromMEM & regwriteWB & (rdWB != 0) & (rdWB == rs2EX);
	assign forwardbEX = ({2{forwardbEXfromMEM}} & 2'b10) | ({2{forwardbEXfromWB}} & 2'b01);

	//for ops in the MEM stage
	//前面一条指令写，后面一条是store指令要存这个写入的数
	assign forwardMEM = (memwriteMEM & regwriteWB & rs2MEM == rdWB);

	//stalls
	//EX阶段指令是lw，ID阶段指令不是sw，那么是一个load-use的冒险
	wire lwstallID = ~memwriteID & memtoregEX & ((rdEX == rs1ID) | (rdEX == rs2ID)); 
	//ID阶段指令是branch指令，EX阶段指令写branch的源操作数，或者MEM阶段是lw指令，读出branch的源操作数 
	wire bstallID  = branchID & ((regwriteEX & ((rdEX == rs1ID) | (rdEX == rs2ID))) 
								  | (memtoregMEM & ((rdMEM == rs1ID) | (rdMEM == rs2ID))));
	//ID阶段是jalr指令，EX阶段指令写jalr的rs1寄存器，或者MEM阶段的指令是lw，读出jalr的rs1寄存器的值
	wire jalrstallID  = jalrID & ((regwriteEX & (rdEX == rs1ID)) | (memtoregMEM & (rdMEM == rs1ID)));

	//stall ID的指令意味着同时stall IF的指令以及新EX的指令为空，即flushE
	assign stallID = lwstallID | bstallID | jalrstallID;
	assign stallIF = stallID;
	assign flushEX = stallID;

	assign flushID = pcsrcID & ~stallID; //如果不stallD，且如果ID阶段指令是无条件跳转，或者条件跳转满足，pcsrc要选择新目标地址，那么就应该flushD，清除pc+4处取出的指令

endmodule
